// Copyright 2023 The Pigweed Authors
//
// Licensed under the Apache License, Version 2.0 (the "License"); you may not
// use this file except in compliance with the License. You may obtain a copy of
// the License at
//
//     https://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
// WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
// License for the specific language governing permissions and limitations under
// the License.

`timescale 1ns / 1ps
//////////////////////////////////////////////////////////////////////////////////
// SPI for INA229
// Refer Spec: https://www.ti.com/lit/ds/symlink/ina229.pdf?ts=1692387535714&ref_url=https%253A%252F%252Fwww.ti.com%252Fproduct%252FINA229
//////////////////////////////////////////////////////////////////////////////////

module spi_m_core #(
    parameter SCK_DIVISOR = 10,
    parameter MCLK_CYCLE_TIME = 10
) (
    input mclk_i,
    input rst_i,
    input r_w_en_i,
    input en_i,
    input [5:0] rwaddr_i,
    input [15:0] wdata_i,
    output reg dn_o,
    output reg [39:0] rdata_o,
    // spi interface
    input miso_i,
    output reg sclk_o,
    output reg mosi_o,
    output reg cs_o
);


  reg
      miso_r0,
      miso_r1,
      miso_r2,
      rising_edge_det,
      falling_edge_det,
      rising_edge_det_r0,
      rising_edge_det_r1,
      rising_edge_det_r2,
      en_r,
      sck_div_ctr_clr,
      sclk_clr;
  reg
      wr_done_r,
      rd_en_r,
      rd_done_r,
      dat_out_ctr_en_r,
      r_w_en_r,
      falling_edge_det_r2,
      falling_edge_det_r1,
      falling_edge_det_r0,
      wait_ctr_en_r,
      spi_en_r,
      dn_r,
      wait_dn_r;
  reg [3:0] sck_div_ctr_r, wait_ctr_r;
  reg [2:0] spi_m_state;
  reg [5:0] rwaddr_latch_r;
  reg [6:0] rwaddr_r, max_bit_count_r, dat_out_ctr_r, dat_in_ctr;
  reg [15:0] wdata_r;
  reg [23:0] dat_out_r;
  reg [39:0] dat_in_r;

  localparam IDLE = 0;  // Power On State
  localparam EN = 1;  // Transfer Enabled. Waiting for Transfer Done.
  localparam DN = 2;  // Transfer Done State. Drive dn_o high.
  localparam WAIT = 3; // State to Add the min. wait time required b/w CS assertion/de-assertion and CLK Enable/Disable.

  localparam CLK_DIV = SCK_DIVISOR - 1;  // Clock Divisor wrt 100MHz mclk.
  localparam CLK_DIV_2 = CLK_DIV >> 1;  // CLK_DIV/2
  localparam t_SCLK_CS_WAIT = 0;  // # no. of mclk cycle wait for t_SCLK_CS

  // SPI State Machine. Mealy, 3 States. Control Signals = en_r, dn_r.

  always @(posedge mclk_i) begin
    if (rst_i) spi_m_state <= IDLE;
    else begin
      case (spi_m_state)
        IDLE: begin
          if (en_r) spi_m_state <= EN;
          else spi_m_state <= IDLE;
        end
        EN: begin
          if (dn_r) spi_m_state <= DN;
          else spi_m_state <= EN;
        end
        DN: begin
          // if (dn_r)
          //	spi_m_state<=DN;
          // else
          spi_m_state <= IDLE;
        end
        default: begin
          spi_m_state <= IDLE;
        end
      endcase
    end
  end

  // State Machine O/Ps.

  always @(posedge mclk_i) begin
    if (rst_i) begin
      cs_o <= 1;
      dat_out_ctr_en_r <= 0;
      spi_en_r <= 0;
      wait_ctr_en_r <= 0;
      dn_o <= 0;
    end else begin
      case (spi_m_state)
        IDLE: begin
          spi_en_r <= 0;
          dat_out_ctr_en_r <= 0;
          wait_ctr_en_r <= 0;
          dn_o <= 0;
          if (en_r) cs_o <= 0;
          else cs_o <= 1;
        end
        EN: begin
          cs_o <= 0;
          wait_ctr_en_r <= 0;
          dn_o <= 0;
          if (dn_r) begin
            spi_en_r <= 0;
            dat_out_ctr_en_r <= 0;
          end else begin
            dat_out_ctr_en_r <= 1;
            spi_en_r <= 1;
          end
        end
        DN: begin
          cs_o <= 0;
          dat_out_ctr_en_r <= 0;
          spi_en_r <= 0;
          wait_ctr_en_r <= 0;
          dn_o <= 1;
        end
      endcase
    end
  end

  // SCK Logic begin
  always @(posedge mclk_i) begin
    if (spi_en_r) begin
      if (sck_div_ctr_clr) sck_div_ctr_r <= 0;
      else sck_div_ctr_r <= sck_div_ctr_r + 1;
    end else sck_div_ctr_r <= 0;
  end

  always @(posedge mclk_i) begin
    if (rst_i) begin
      sclk_o <= 0;
    end else begin
      sclk_o <= (spi_en_r && (sck_div_ctr_r > CLK_DIV_2));  // default state = 0
    end
  end
  // SCK Logic End


  // Data Transmit and Receive Logic Begin
  // Pulses to sync transmit and capture with respective clock edges.
  always @(posedge mclk_i) begin
    rising_edge_det  <= (sck_div_ctr_r == CLK_DIV_2);
    falling_edge_det <= (sck_div_ctr_r == CLK_DIV);
  end

  always @(posedge mclk_i) begin
    if (dat_out_ctr_en_r) begin
      if (falling_edge_det) dat_out_ctr_r <= dat_out_ctr_r + 1;
    end else dat_out_ctr_r <= 0;
  end

  always @(posedge mclk_i) begin
    if (rst_i) en_r <= 0;
    else begin
      if (dn_r) en_r <= 0;
      else if (en_i) en_r <= 1;
    end
  end

  always @(posedge mclk_i) begin
    if (rst_i) begin
      rwaddr_r <= 0;
      rwaddr_latch_r <= 0;
      wdata_r <= 0;
      r_w_en_r <= 0;
      dat_out_r <= 0;
    end else begin
      if (en_i) begin
        dat_out_r <= {rwaddr_i, 1'b0, r_w_en_i, wdata_i};
        rwaddr_r <= {rwaddr_i, 1'b0};
        rwaddr_latch_r <= rwaddr_i;
        wdata_r <= wdata_i;
        r_w_en_r <= r_w_en_i;
      end
    end
  end

  always @(posedge mclk_i) begin
    if (r_w_en_r) begin
      if (rwaddr_latch_r== 6'd4 || rwaddr_latch_r== 6'd5 || rwaddr_latch_r== 6'd7 || rwaddr_latch_r== 6'd8)
        max_bit_count_r <= 24;
      else if (rwaddr_latch_r == 6'd9 || rwaddr_latch_r == 6'd10) max_bit_count_r <= 40;
      else max_bit_count_r <= 16;
    end else begin
      max_bit_count_r <= 16;
    end
  end

  always @(posedge mclk_i) begin
    wr_done_r<=((!r_w_en_r) & (dat_out_ctr_r==24));// && falling_edge_det_r1); // added sync with falling_edge_det_r2
    rd_en_r <= ((dat_out_ctr_r > 8) && r_w_en_r);  // changed from (dat_out_ctr_r>=8)
    rd_done_r <= (dat_in_ctr == (max_bit_count_r - 1));
    wait_dn_r <= (wait_ctr_r == t_SCLK_CS_WAIT);
    sck_div_ctr_clr <= (sck_div_ctr_r == (CLK_DIV - 1));
  end

  always @(posedge mclk_i) begin
    if (rst_i) begin
      mosi_o <= 1;
    end else begin
      if (spi_en_r) begin
        if (rising_edge_det) begin
          if (rd_en_r) mosi_o <= 0;
          else mosi_o <= dat_out_r[23-dat_out_ctr_r];
        end
      end else
        mosi_o <= 1;  // if (cs_o) mosi_o<=1; // change-might create a crit path b/w cs_o and mosi
    end
  end

  always @(posedge mclk_i) begin
    if (rd_en_r) begin
      if (falling_edge_det_r0) begin
        dat_in_ctr <= dat_in_ctr + 1;
      end
    end else begin
      dat_in_ctr <= 0;
    end
  end

  always @(posedge mclk_i) begin
    if (rd_en_r) begin
      if (falling_edge_det_r1) begin
        dat_in_r <= {dat_in_r, miso_r1};
      end
    end else begin
      dat_in_r <= 0;
    end
  end

  always @(posedge mclk_i) begin
    if (rst_i)  // Rst not required.
      rdata_o <= 0;
    else if (rd_done_r & rising_edge_det) rdata_o <= dat_in_r;
  end

  always @(posedge mclk_i) begin
    if (wait_ctr_en_r) wait_ctr_r <= wait_ctr_r + 1;
    else wait_ctr_r <= 0;
  end


  // Using shift reg. pattern. Unsure of ASYNC_REG placement on icestorm.
  always @(posedge mclk_i) begin
    if (rst_i) begin
      {miso_r0, miso_r1, miso_r2} <= 0;
      {rising_edge_det_r0, rising_edge_det_r1, rising_edge_det_r2} <= 0;
      {falling_edge_det_r0, falling_edge_det_r1, falling_edge_det_r2} <= 0;
    end else begin
      {miso_r2, miso_r1, miso_r0} <= {miso_r1, miso_r0, miso_i};
      {rising_edge_det_r2, rising_edge_det_r1, rising_edge_det_r0} <= {
        rising_edge_det_r1, rising_edge_det_r0, rising_edge_det
      };
      {falling_edge_det_r2, falling_edge_det_r1, falling_edge_det_r0} <= {
        falling_edge_det_r1, falling_edge_det_r0, falling_edge_det
      };
    end
  end

  always @(posedge mclk_i) begin
    dn_r <= (wr_done_r | (rd_done_r));
  end

  // Data Transmit and Receive Logic End
endmodule
